//-*-c++-*------------------------------------------------------------
//
// File name : bioArithDivide.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Tue Apr 28 20:58:50 2009
//
//--------------------------------------------------------------------

#include <sstream>
#include "patMath.h"
#include "patDisplay.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "bioLiteralRepository.h"
#include "bioArithDivide.h"
#include "bioArithMult.h"
#include "bioArithUnaryMinus.h"
#include "bioArithBinaryMinus.h"
#include "bioArithPower.h"
#include "bioArithConstant.h"
#include "bioArithLiteral.h"
#include "bioArithNamedExpression.h"
#include "bioExpressionRepository.h"
#include "bioArithCompositeLiteral.h"

bioArithDivide::bioArithDivide(bioExpressionRepository* rep,
			       patULong par,
                               patULong left, 
                               patULong right,
patError*& err) : 
  bioArithBinaryExpression(rep,par,left,right,err) {
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

}

bioArithDivide::~bioArithDivide() {}

patString bioArithDivide::getOperatorName() const {
  return ("/") ;
}

patReal bioArithDivide::getValue(patBoolean prepareGradient, patError*& err)   {
#ifndef BIO_NOCHECK
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patReal();
  }
#endif
  
  patReal left ;
  left = leftChild->getValue(prepareGradient,err) ;

#ifndef BIO_NOCHECK
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
#endif

  if (left == 0.0) {
    return 0.0 ;
  }
  patReal right ;
  right = rightChild->getValue(prepareGradient,err) ;
#ifndef BIO_NOCHECK
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
#endif
  if (right == 0.0) {
    return patMaxReal ;
  }

  patReal result = left / right ;

#ifndef BIO_NOCHECK
  if (!finite(result)) {
    return patMaxReal ;
  }
  else {
    return result ;
  }
#else
  return result ;
#endif
}

bioExpression* bioArithDivide::getDerivative(patULong aLiteralId, 
					     patError*& err) const {

  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return NULL;
  }
  
  bioExpression* leftValue = leftChild->getDeepCopy(theRepository,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  bioExpression* rightValue = rightChild->getDeepCopy(theRepository,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  bioExpression* leftDeriv = leftChild->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  bioExpression* rightDeriv = rightChild->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  if (leftChild->dependsOf(aLiteralId)) {
    if (rightChild->dependsOf(aLiteralId)) {
      // THIS WILL HAVE TO BE REIMPLMENTED WITH THE NEW DEFINITION OF SEQUENCES
//       patULong lId = bioLiteralRepository::the()->addCompositeLiteral(err) ;
//       if (err != NULL) {
// 	WARNING(err->describe()) ;
//       }
      
      
//       bioArithLiteral* theSimplifiedExpression =  new bioArithLiteral(NULL,theId) ;
//       // Left variable, right variable
//       bioExpression* l = new bioArithMult(NULL,theSimplifiedExpression,leftDeriv) ;
//       bioExpression* r = new bioArithMult(NULL,leftValue,rightDeriv) ;
//       bioExpression* n = new bioArithBinaryMinus(NULL,l,r) ;
//       bioExpression* d = new bioArithMult(NULL,theSimplifiedExpression,theSimplifiedExpression) ;
//       bioExpression* ratio = new bioArithDivide(NULL,n,d) ;
//       bioArithSequence* result = new bioArithSequence(NULL,rightValue,ratio,theId,err) ; 
      bioExpression* lderiv_r = new bioArithMult(theRepository,patBadId,leftDeriv->getId(),rightValue->getId(),err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return NULL ;
	}
	bioExpression* rderiv_l = new bioArithMult(theRepository,patBadId,rightDeriv->getId(),leftValue->getId(),err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return NULL ;
	}
	bioExpression* numerator = new bioArithBinaryMinus(theRepository,patBadId,lderiv_r->getId(),rderiv_l->getId(),err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return NULL ;
	}
	bioExpression* denominator = new bioArithMult(theRepository,patBadId,rightValue->getId(),rightValue->getId(),err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return NULL ;
	}
	bioExpression* result = new bioArithDivide(theRepository,patBadId,numerator->getId(),denominator->getId(),err);
      if (err != NULL) {
	WARNING(err->describe()) ;
	return NULL;
      }
      return result ;
    }
    else {
      // Left variable, right constant
      bioExpression* l = new bioArithDivide(theRepository,patBadId,leftDeriv->getId(),rightValue->getId(),err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return NULL ;
	}
      return l ;
    }
  }
  else {
    if (rightChild->dependsOf(aLiteralId)) {
      //Left constant, right variable
      bioExpression* r = new bioArithMult(theRepository,patBadId,leftValue->getId(),rightDeriv->getId(),err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return NULL;
	}
	bioExpression* n = new bioArithUnaryMinus(theRepository,patBadId,r->getId(),err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return NULL ;
	}
	bioExpression* d = new bioArithMult(theRepository,patBadId,rightValue->getId(),rightValue->getId(),err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return NULL ;
	}
	bioExpression* result = new bioArithDivide(theRepository,patBadId,n->getId(),d->getId(),err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return NULL ;
	}

      return result ;
    }
    else {
      //Left constant, right constant
      bioExpression* result = new bioArithConstant(theRepository,patBadId,0) ;
      return result ;
    }
  }

  
  //  patReal result = (rightValue * leftDeriv - leftValue * rightDeriv) / 
  //                 (rightValue * rightValue) ;

}

bioArithDivide* bioArithDivide::getDeepCopy(bioExpressionRepository* rep,
					    patError*& err) const {
  bioExpression* leftClone(NULL) ;
  bioExpression* rightClone(NULL) ;
  if (leftChild != NULL) {
    leftClone = leftChild->getDeepCopy(rep,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (leftClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }
  if (rightChild != NULL) {
    rightClone = rightChild->getDeepCopy(rep,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (rightClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }

  bioArithDivide* theNode = 
    new bioArithDivide(rep,patBadId,leftClone->getId(),rightClone->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  return theNode ;
}

bioArithDivide* bioArithDivide::getShallowCopy(bioExpressionRepository* rep,
					    patError*& err) const {
  bioArithDivide* theNode = 
    new bioArithDivide(rep,patBadId,leftChild->getId(),rightChild->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  return theNode ;
}

patString bioArithDivide::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patBoolean lSimple = leftChild->isSimple() ;
  patString vLeft ;
  if (lSimple) {
    vLeft = leftChild->getSimpleCppCode(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  else {
    vLeft = leftChild->getCppCode(prefix, err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  patBoolean rSimple = rightChild->isSimple() ;
  patString vRight ;
  if (rSimple) {
    vRight = rightChild->getSimpleCppCode(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  else {
    vRight = rightChild->getCppCode(prefix, err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  stringstream str ;
  str << getCppResultDeclaration(prefix) << ";" ;
  if (!lSimple) {
    str << vLeft << endl ;
  }
  if (!rSimple) {
    str << vRight << endl ;
  }
  if (!lSimple) {
    str << "if (" <<leftChild->getCppResult(prefix) << "== 0) {" << endl ;
    str << "    " << getCppResult(prefix) << "= 0 ;" << endl ;
    str << "}" << endl ;
  }
  else {
    str << "if ((" << vLeft << ") == 0) {" << endl ;
    str << "    " << getCppResult(prefix) << "= 0 ;" << endl ;
    str << "}" << endl ;
  }
  str << " else {" << endl ;
  str << getCppResult(prefix) << "= (" ;
  if (lSimple) {
    str << vLeft ;
  }
  else {
    str << leftChild->getCppResult(prefix) ;
  }
  str << " ) / ( " ;
  if (rSimple) {
    str << vRight ;
  }
  else {
    str << rightChild->getCppResult(prefix) ;
  }
  str << ");" << endl ;
  str << "}" << endl ;
  str << "    if (!finite(" << getCppResult(prefix) << ")) {" << endl ;
  str << "      " << getCppResult(prefix) << " =  patMaxReal ;" << endl ;
  str << "    }" << endl ;
  //  str << "DEBUG_MESSAGE(\"" << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
  codeMustBeGenerated = patFALSE ;
  return patString(str.str()) ;
  }
  else {
    return patString() ;
  }
}


patString bioArithDivide::getSimpleCppCode(patError*& err) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patString vLeft = leftChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  patString vRight = rightChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }

  if (vLeft == "0.0") {
    return patString("0.0") ;
  }

  stringstream str ;
  str << "(((" << vLeft << ")==0)?0:((" << vLeft << ")/(" << vRight << ")))"  ;
  return patString(str.str()) ;
}



patBoolean bioArithDivide::isSimple() const {
  return patFALSE ;
}

patBoolean bioArithDivide::isStructurallyZero() const {
  return (leftChild->isStructurallyZero()) ;
}

patString bioArithDivide::getExpressionString() const {
  stringstream str ;
  str << '/' ;
  if (leftChild != NULL) {
    str << '{' << leftChild->getExpressionString() << '}' ;
  }
  if (rightChild != NULL) {
    str << '{' << rightChild->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;

}



  

bioFunctionAndDerivatives* bioArithDivide::getNumericalFunctionAndGradient(vector<patULong> literalIds,
							   patError*& err) {
  if (result.empty()) {
    result.resize(literalIds.size()) ;
  }
  bioFunctionAndDerivatives* l = leftChild->getNumericalFunctionAndGradient(literalIds,err) ;
#ifndef BIO_NOCHECK
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
#endif
  if (l->theFunction == 0.0) {
    patReal r = rightChild->getValue(patFALSE,err) ;
    // l = 0
    if (r  == 0.0) {
      // l= 0, r = 0
      result.theFunction =  0.0 ;
      for (patULong i = 0 ; i < literalIds.size() ; ++i) {
	result.theGradient[i] = 0.0 ;
      }
    } 
    else if (r == 1.0) {
      // l = 0, r = 1
      result.theFunction = 0.0 ;
      for (patULong i = 0 ; i < literalIds.size() ; ++i) {
	result.theGradient[i] = l->theGradient[i] ;
      }
    }
    else {
      // l=0, r != 0, r != 1
      result.theFunction =  0.0 ;
      for (patULong i = 0 ; i < literalIds.size() ; ++i) {
	if (l->theGradient[i] == 0) {
	  result.theGradient[i] = 0.0 ;
	}
	else {
	  result.theGradient[i] = l->theGradient[i] / r ;
	}
      }
    }
  }
  else {
    bioFunctionAndDerivatives* r = 
      rightChild->getNumericalFunctionAndGradient(literalIds,err) ;
#ifndef BIO_NOCHECK
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
#endif
    // l != 0
    if (r->theFunction == 0.0) {
      // l != 0, r = 0
      result.theFunction =  patMaxReal ;
      for (patULong i = 0 ; i < literalIds.size() ; ++i) {
	result.theGradient[i] = patMaxReal ;
      }
    } 
    else if (r->theFunction == 1.0) {
      // l != 0, r = 1
      result.theFunction =  l->theFunction ;
      for (patULong i = 0 ; i < literalIds.size() ; ++i) {
	if (l->theGradient[i] == 0.0) {
	  if (r->theGradient[i] == 0.0) {
	    result.theGradient[i] = 0.0 ;
	  } 
	  else {
	    result.theGradient[i] = - r->theGradient[i] * l->theFunction ;
	  }
	}
	else {
	  if (r->theGradient[i] == 0.0) {
	    result.theGradient[i] = l->theGradient[i] ;
	  } 
	  else {
	    result.theGradient[i] = l->theGradient[i] - r->theGradient[i] * l->theFunction ;
	  }
	}
      }
    }
    else {
      // l != 0, r != 0, r != 1
      result.theFunction =  l->theFunction / r->theFunction ;
      for (patULong i = 0 ; i < literalIds.size() ; ++i) {
	result.theGradient[i] = (l->theGradient[i] * r->theFunction -
				 r->theGradient[i] * l->theFunction) 
	  / (r->theFunction * r->theFunction) ;
      }
    }
    
  }
#ifndef BIO_NOCHECK
  for (patULong i = 0 ; i < literalIds.size() ; ++i) {
    if (!finite(result.theGradient[i])) {
      result.theGradient[i] = patMaxReal ;
    }
  }
#endif  
  return &result ;
}

