//-*-c++-*------------------------------------------------------------
//
// File name : bioArithUnaryExpression.cc
// Author :    Michel Bierlaire
// Date :      Wed Apr 13 15:28:44 2011
//
//--------------------------------------------------------------------

#include "patDisplay.h"
#include "bioArithUnaryExpression.h"
#include "patErrNullPointer.h"
#include "bioArithConstant.h"
#include "bioExpressionRepository.h"
#include "patErrMiscError.h"

bioArithUnaryExpression::bioArithUnaryExpression(bioExpressionRepository* rep,
						 patULong par,
						 patULong c,
						 patError*& err) :
  bioExpression(rep,par), childId(c) {

  child = theRepository->getExpression(childId) ;
  if (child == NULL) {
    stringstream str ;
    str << "No expression with id " << childId ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return ;
  }
  child->setParent(getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

}

bioArithUnaryExpression::~bioArithUnaryExpression() {

}

bioExpression* bioArithUnaryExpression::getChild() const {
  return child ;
}

patString bioArithUnaryExpression::getExpression(patError*& err) const {
  if (child == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patString childResult = child->getExpression(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  
  patString result =  getOperatorName() ;
  result += '(' ;
  result += childResult ;
  result += ')' ;
  return result ;
}

patULong bioArithUnaryExpression::getNumberOfOperations() const {
  if (child == NULL) {
    return 0 ;
  }
  patULong result = child->getNumberOfOperations() ;
  return (1 + result) ;
}

patBoolean bioArithUnaryExpression::dependsOf(patULong aLiteralId) const {
  if (child == NULL) {
    return patFALSE ;
  }  
  return (child->dependsOf(aLiteralId)) ;
}

void bioArithUnaryExpression::forceCodeGeneration() {
  codeMustBeGenerated = patTRUE ;
  if (child != NULL) {
    child->forceCodeGeneration() ;
  }
}

patBoolean bioArithUnaryExpression::containsAnIterator() const {
  return child->containsAnIterator() ;
}

patBoolean bioArithUnaryExpression::containsAnIntegral() const {
  return child->containsAnIntegral() ;
}

patBoolean bioArithUnaryExpression::containsASequence() const {
  return child->containsASequence() ;
}

void bioArithUnaryExpression::simplifyZeros(patError*& err) {
  bioExpression* newChild(NULL) ;
  if (child != NULL) {
    if (child->isStructurallyZero()) {
      newChild = theRepository->getZero() ;
    }
    else if (child->isConstant()) {
      patReal value = child->getValue(patFALSE,err) ;
      if (err != NULL) {
        WARNING(err->describe()) ;
        return ;
      }
      newChild = new bioArithConstant(theRepository,patBadId,value) ;
    }
    else {
      child->simplifyZeros(err) ;
      if (err != NULL) {
        WARNING(err->describe()) ;
        return ;
      }
    }
  }
  if (newChild != NULL) {
    child = newChild ;
    childId = child->getId() ;
  }
}

void bioArithUnaryExpression::generateGslFunctions(ostream& str_h,patString prefix,patError*& err) {

  if (child != NULL) {
    child->generateGslFunctions(str_h,prefix,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
  }
}

patString bioArithUnaryExpression::getIncludeStatements(patString prefix) const {
  if (child != NULL) {
    child->getIncludeStatements(prefix) ;
  }
  return patString() ;
}

void bioArithUnaryExpression::setVariables(const patVariables* x) {
  __x = x ;
  child->setVariables(x) ;
  for (vector<bioExpression*>::iterator i = relatedExpressions.begin() ;
       i != relatedExpressions.end() ;
       ++i) {
    (*i)->setVariables(x) ;
  }
}

void bioArithUnaryExpression::setDraws(const vector<vector< patReal > >* d) {
  __draws = d ;
  child->setDraws(d) ;
  for (vector<bioExpression*>::iterator i = relatedExpressions.begin() ;
       i != relatedExpressions.end() ;
       ++i) {
    (*i)->setDraws(d) ;
  }
}

patBoolean bioArithUnaryExpression::isSimple() const {
  return child->isSimple() ; 
}

void bioArithUnaryExpression::setSample(const bioSample* s) {
  theSample = s ;
  child->setSample(s) ;
  for (vector<bioExpression*>::iterator i = relatedExpressions.begin() ;
       i != relatedExpressions.end() ;
       ++i) {
    (*i)->setSample(s) ;
  }

}

void bioArithUnaryExpression::setCurrentSpan(bioIteratorSpan aSpan) {
  theCurrentSpan = aSpan ;
  child->setCurrentSpan(aSpan) ;
  for (vector<bioExpression*>::iterator i = relatedExpressions.begin() ;
       i != relatedExpressions.end() ;
       ++i) {
    (*i)->setCurrentSpan(aSpan) ;
  }
}
void bioArithUnaryExpression::setThreadSpan(bioIteratorSpan aSpan) {
  theThreadSpan = aSpan ;
  child->setThreadSpan(aSpan) ;
  for (vector<bioExpression*>::iterator i = relatedExpressions.begin() ;
       i != relatedExpressions.end() ;
       ++i) {
    (*i)->setThreadSpan(aSpan) ;
  }
}

void bioArithUnaryExpression::collectExpressionIds(set<patULong>* s) const {
  s->insert(getId()) ;
  child->collectExpressionIds(s) ;
  for (vector<bioExpression*>::const_iterator i = relatedExpressions.begin() ;
       i != relatedExpressions.end() ;
       ++i) {
    (*i)->collectExpressionIds(s) ;
  }
}



patString bioArithUnaryExpression::check(patError*& err) const {
  stringstream str ;
  if (child->getId() != childId) {
    str << "Incompatible IDS for children: " << child->getId() << " and " << childId ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return patString() ;
  }
  if (child->getParent() != getId()) {
    str << "Incompatible IDS for parent: " << child->getParent() << " and " << getId() ;
    str << "Expression: " << *this << endl ;
    str << "Child: " << *child << endl ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return patString() ;
  }
  return patString() ;
}
