//-*-c++-*------------------------------------------------------------
//
// File name : bioMain.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Fri Jul 17 16:41:28 2009
//
//--------------------------------------------------------------------

#include <Python.h>
#include <iomanip>
#include "bioExpressionRepository.h"
#include "bioMinimizationProblem.h"
#include "patErrNullPointer.h"
#include "patTimeInterval.h"
#include "bioMain.h"
#include "bioRandomDraws.h"
#include "bioParameters.h"
#include "bioSample.h"
#include "bioRowIterator.h"
#include "patDisplay.h"
#include "patFileNames.h"
#include "bioVersion.h"
#include "bioModelParser.h"
#include "trBounds.h"
#include "trSimpleBoundsAlgo.h"
#include "patCfsqp.h"
#include "patSolvOpt.h"
#include "patDonlp2.h"
#include "bioRawResults.h"
#include "bioReporting.h"
#include "bioStatistics.h"
#include "bioSimulation.h"
#include "bioConstraints.h"
#include "bioIterationBackup.h"
#include "bioPrecompiledFunction.h"
#include "bioArithPrint.h"

bioMain::bioMain() : theFunction(NULL),theSample(NULL), theModel(NULL), theStatistics(NULL), theConstraints(NULL), theRawResults(NULL), theRepository(NULL) {

}

void bioMain::setFunction(bioPythonWrapper* f) {
  theFunction = f ;
}

void bioMain::setStatistics(bioStatistics* s) {
  theStatistics = s ;
}

void bioMain::setConstraints(bioConstraints* c) {
  theConstraints= c ;
}

void bioMain::run(int argc, char *argv[]) {

  patError* err(NULL) ;
  

  // Reads the sample file

  patString sampleFile("sample.dat") ;
  if (argc >= 3) {
    sampleFile = patString(patString(argv[2])) ;
  }
    DEBUG_MESSAGE("Sample file: " << sampleFile) ;
  
  theSample = new bioSample(sampleFile,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    exit(-1) ;
  }


  // Init Python
  Py_Initialize() ;

  PyRun_SimpleString("import sys") ;
  PyRun_SimpleString("sys.path.append('./module/')") ;



  // Get model name


  patString fn(argv[1]) ;

  //  DEBUG_MESSAGE("Model file: " << argv[1]) ;
  patString pythonFileExtension(".py") ;
  int extensionPos = fn.size()-pythonFileExtension.size() ;

  // Remove .py extension if present
  if (fn.find(pythonFileExtension, extensionPos)!=string::npos) {
    fn.erase(extensionPos) ;
  }

  patFileNames::the()->setModelName(fn) ;

  GENERAL_MESSAGE("READ MODEL FROM " << fn) ;
  bioModelParser* parser = new bioModelParser(fn) ;
  parser->setSampleFile(sampleFile) ;
  theModel = parser->readModel(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    exit(-1) ;
  }
  theRepository = theModel->getRepository() ;
  if (theRepository == NULL) {
    err = new patErrNullPointer("bioExpressionRepository") ;
    WARNING(err->describe()) ;
    exit(-1) ;
  }
  GENERAL_MESSAGE("MODEL READ") ;
  bioExpression* excludeExpression = theRepository->getExpression(theModel->getExcludeExpr()) ;
  theSample->prepare(excludeExpression,err) ;
  if (err != NULL) {
    WARNING(err->describe()); 
    exit(-1) ;
  }

  setParameters(err) ;
  if (err != NULL) {
    WARNING(err->describe()); 
    exit(-1) ;
  }

  bioLiteralRepository::the()->prepareMemory();


  theReport.computeFromSample(theSample,err) ;

  DEBUG_MESSAGE("Sample of size: " << theSample->size()) ;


  ofstream po("__parametersUsed.py") ;
  po << bioParameters::the()->printPythonCode() << endl;
  po.close() ;

  bioRandomDraws::the()->populateListOfIds(theSample,err) ;
  if (err != NULL) {
    WARNING(err->describe()); 
    exit(-1) ;
  }



  bioRandomDraws::the()->generateDraws(err) ;
  if (err != NULL) {
    WARNING(err->describe()); 
    exit(-1) ;
  }


  //  DEBUG_MESSAGE(*bioRandomDraws::the()) ;

  map<patString, patULong>* stats = theModel->getStatistics() ;

  bioIteratorSpan __theThreadSpan ;

  if (stats != NULL) {
    for (map<patString, patULong>::iterator i = stats->begin() ;
	 i != stats->end() ;
	 ++i) {
      bioExpression* expr = theRepository->getExpression(i->second) ;
      if (expr == NULL) {
	err = new patErrNullPointer("bioExpression") ;
	WARNING(err->describe()) ;
	return ;
      }
      expr->setThreadSpan(__theThreadSpan) ;
      expr->setSample(theSample) ;
      patReal result = expr->getValue(patFALSE,err) ;
      if (err != NULL) {
	WARNING(err->describe()); 
	return ;
      }
      theReport.addStatistic(i->first,result) ;
    }

  }
  //  statistics(err) ;





  if (theModel->mustEstimate()) {
    bioExpressionRepository* theRepository = theModel->getRepository();
    patULong exprId = theModel->getFormula() ;
    theFunction = new bioPrecompiledFunction(theRepository,exprId,err) ;
    if (err != NULL) {
      WARNING(err->describe()); 
      exit(-1) ;
    }
    estimate(err) ;
    if (err != NULL) {
      WARNING(err->describe()); 
      exit(-1) ;
    }
  }

  if (theModel->mustSimulate()) {

    //For simulation, only one thread is used
    int nt = bioParameters::the()->getValueInt("numberOfThreads",err) ; 
    if (err != NULL) {
      WARNING(err->describe()); 
      exit(-1) ;
    }
    if (nt > 1) {
      WARNING("Only one thread is used for simulation."); 
      bioParameters::the()->setValueInt("numberOfThreads",1,err) ; 
    }
    simulate(err) ;
    if (err != NULL) {
      WARNING(err->describe()); 
      exit(-1) ;
    }

  }
  
  patString htmlFile = patFileNames::the()->getHtmlFile(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  
  theReport.writeHtml(htmlFile,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

  patString latexFile = patFileNames::the()->getLatexFile(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

  theReport.writeLaTeX(latexFile,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

}

void bioMain::estimate(patError*& err) {

  if (theSample == NULL) {
    err = new patErrMiscError("No sample has been loaded. Biogeme cannot estimate the parameters") ;
    WARNING(err->describe()) ;
    return ;
  }

  theFunction->setSample(theSample,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }


  // Bounds

  patVariables l = bioLiteralRepository::the()->getLowerBounds() ;
  patVariables u = bioLiteralRepository::the()->getUpperBounds() ;
  trBounds theBounds(l,u) ;
  
  patVariables beta = bioLiteralRepository::the()->getBetaValues(patFALSE) ;
  DEBUG_MESSAGE("Number of betas = " << beta.size()) ;
  patBoolean success ;
  patReal initLikelihood(0.0) ;
  patBoolean computeInit = (bioParameters::the()->getValueInt("computeInitLoglikelihood",err) != 0) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  
  patAbsTime beg ;
  patAbsTime endc ;

  
  if (computeInit) {
    beg.setTimeOfDay() ;
    initLikelihood = -theFunction->getFunction(&beta,&success,err) ;
    endc.setTimeOfDay() ;
    patTimeInterval til0(beg,endc) ;
    GENERAL_MESSAGE("Init. log-likelihood: " << initLikelihood << " ["<<til0.getLength()<<"]") ;
  }
  patBoolean checkDeriv = (bioParameters::the()->getValueInt("checkDerivatives",err) != 0) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

  // WARNING("Derivatives are always checked") ;
  // checkDeriv = patTRUE ;

  if (checkDeriv) {
    patVariables grad(beta.size(),0.0) ;
    beg.setTimeOfDay() ;
    theFunction->getGradient(&beta,&grad,&success,err) ;
    endc.setTimeOfDay() ;
    patTimeInterval tig(beg,endc) ;
    patVariables finDiffGrad(beta.size(),0.0) ;
    beg.setTimeOfDay() ;
    theFunction->getFinDiffGradient(&beta,&finDiffGrad,&success,err) ;
    endc.setTimeOfDay() ;
    patTimeInterval tigfd(beg,endc) ;

    
    GENERAL_MESSAGE("Analytical\t\tFin. diff.\tError\tName") ;
    for (patULong i = 0 ; i < grad.size() ; ++i) {
      patString name = bioLiteralRepository::the()->getBetaName(i,patFALSE,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
      GENERAL_MESSAGE( setprecision(7) << setiosflags(ios::scientific|ios::showpos) << grad[i] << '\t' << finDiffGrad[i] << '\t' << setprecision(2) << 100*(grad[i]-finDiffGrad[i])/finDiffGrad[i]<<"% \t" << name) ;
    }

  }


  // exit(0) ;
  // WARNING("Exit for debug") ;
  //Optimization problem

  bioMinimizationProblem theProblem(theFunction, theBounds,theTrParameters) ;
  

  // Constraints
  if (theConstraints != NULL) {
    theConstraints->addConstraints() ;
    vector<bioConstraintWrapper*>* c = theConstraints->getConstraints() ;
    for (vector<bioConstraintWrapper*>::iterator i = c->begin() ;
	 i != c->end() ;
	 ++i) {
      theProblem.addEqualityConstraint(*i) ;
    }
    
  }
  
  // Optimization algorithm

  beg.setTimeOfDay() ;

  patString selectedAlgo = bioParameters::the()->getValueString("optimizationAlgorithm",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

  if (theProblem.nNonTrivialConstraints() > 0 && selectedAlgo == "BIO") {
    if (patCfsqp::isAvailable) {
      selectedAlgo = "CFSQP" ;
      GENERAL_MESSAGE("The problem has " << theProblem.nNonTrivialConstraints() << " constraints. Algorithm BIO cannot be used. CFSQP is used instead.") ; 
    }
    else {
      selectedAlgo = "DONLP2" ;
      GENERAL_MESSAGE("The problem has " << theProblem.nNonTrivialConstraints() << " constraints. Algorithm BIO cannot be used. DONLP2 is used instead.") ; 
    }

    patULong nle = theProblem.nNonLinearEq() ;
    for (patULong i = 0 ; i < nle ; ++i) {
      trFunction* f = theProblem.getNonLinEquality(i,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
      patBoolean success ;
      f->getFunction(&beta,&success,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
    }
  }


  if (!theProblem.isFeasible(beta,err)) {
    WARNING("Starting point not feasible") ;
  }

  trNonLinearAlgo* algo ;
  if (selectedAlgo == "BIO") {
    patIterationBackup* aBackup = new bioIterationBackup() ;
    algo = new trSimpleBoundsAlgo(&theProblem,
				  beta,
				  theTrParameters,
				  aBackup,
				  err) ;
    
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
  }

  else if (selectedAlgo == "CFSQP") {
      
    DETAILED_MESSAGE("Use CFSQP") ;
      
    patIterationBackup* aBackup = new bioIterationBackup() ;
    patCfsqp* algoCfsqp = new patCfsqp(aBackup,&theProblem) ;

    int mode = bioParameters::the()->getValueInt("cfsqpMode",err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    int iprint = bioParameters::the()->getValueInt("cfsqpIprint",err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    int miter = bioParameters::the()->getValueInt("cfsqpMaxIter",err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    double eps = bioParameters::the()->getValueReal("cfsqpEps",err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    double epseqn = bioParameters::the()->getValueReal("cfsqpEpsEqn",err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    double udelta = bioParameters::the()->getValueReal("cfsqpUdelta",err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    patString stopFile = bioParameters::the()->getValueString("stopFileName",err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    
    algoCfsqp->setParameters(mode,
			     iprint,
			     miter,
			     eps,
			     epseqn,
			     udelta,
			     stopFile) ;
    algo = algoCfsqp ;
    if (algo == NULL) {
      err = new patErrNullPointer("patCfsqp") ;
      return ;
    }
    algo->defineStartingPoint(beta) ;
  }
  else if (selectedAlgo == "SOLVOPT") {
    patIterationBackup* aBackup = new bioIterationBackup() ;
    algo = new patSolvOpt(theSolvoptParameters,&theProblem) ;
    if (algo == NULL) {
      err = new patErrNullPointer("patSolvOpt") ;
      WARNING(err->describe()) ;
      return ;
    }
    algo->setBackup(aBackup) ;
    algo->defineStartingPoint(beta) ;
  }
  else if (selectedAlgo == "DONLP2") {
    patIterationBackup* aBackup = new bioIterationBackup() ;
    algo = new patDonlp2(theDonlp2Parameters,&theProblem) ;
    if (algo == NULL) {
      err = new patErrNullPointer("patDonlp2") ;
      WARNING(err->describe()) ;
      return ;     
    }
    algo->setBackup(aBackup) ;
    algo->defineStartingPoint(beta) ;
  }
  else {
    stringstream str ;
    str << "Unknown algorithm: " << selectedAlgo ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return ;
  }

  patString algoName(algo->getName()) ;
  
  DEBUG_MESSAGE("About to run algorithm: " << algoName) ;
  
  patString diag = algo->run(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  endc.setTimeOfDay() ;
  patTimeInterval ti0(beg,endc) ;
  
  DETAILED_MESSAGE("Run time interval: " << ti0 ) ;
  DETAILED_MESSAGE("Run time: " << ti0.getLength()) ;

  patVariables solution = algo->getSolution(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  DEBUG_MESSAGE("Solution: " << solution) ;
  bioLiteralRepository::the()->setBetaValues(solution,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

  patString pythonEstimated = patFileNames::the()->getPythonEstimated(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

  ofstream pyEst(pythonEstimated.c_str()) ;
  pyEst << "# This file has automatically been generated" << endl ; 
  pyEst << "# " << bioVersion::the()->getVersionInfoDate() << endl ;
  pyEst << "# " << bioVersion::the()->getVersionInfoAuthor() << endl ;
  patAbsTime now ;
  now.setTimeOfDay() ;
  pyEst << "# " << now.getTimeString(patTsfFULL) << "</p>" << endl ;
  pyEst << "#" << endl ;
  pyEst << bioLiteralRepository::the()->printFixedParameters(patTRUE) << endl ;
//   pyEst << "# Code to use for simulation" << endl ;
//   pyEst << bioLiteralRepository::the()->printFixedParameters(patFALSE) << endl ;
  pyEst.close() ;
  

  GENERAL_MESSAGE("Python code for the estimated parameters: " << endl << bioLiteralRepository::the()->printFixedParameters()) ;
  GENERAL_MESSAGE("File " << pythonEstimated << " created") ;

  theRawResults = new bioRawResults(&theProblem,solution,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

  theReport.computeEstimationResults(theRawResults,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }


//   // Statistics
  
//   vector<pair<patString,patReal> > stats = theModel->getStatistics() ;
//   if (err != NULL) {
//     WARNING(err->describe()); 
//     return ;
//   }
//   for (vector<pair<patString,patReal> >::iterator i = stats.begin() ;
//        i != stats.end() ;
//        ++i) {
//     theReport.addStatistic(i->first,i->second) ;
//   }
  
  theReport.setDiagnostic(diag) ;
  theReport.setIterations(algo->nbrIter()) ;
  theReport.setRunTime(ti0.getLength()) ;
  theReport.setInitLL(initLikelihood) ;
}


void bioMain::simulate(patError*& err) {

  bioArithPrint* expr = theModel->getSimulation() ;

  if (theSample == NULL) {
    err = new patErrMiscError("No sample has been loaded. Biogeme cannot estimate the parameters") ;
    WARNING(err->describe()) ;
    return ;
  }

  expr->setSample(theSample) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }


  //  patVariables beta = bioLiteralRepository::the()->getBetaValues(patFALSE) ;
  //patBoolean success ;
  //patReal result = theFunction->getFunction(&beta,&success,err) ;
  expr->simulate(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  
  theReport.setSimulationResults(expr->getSimulationResults()) ;

}


void bioMain::statistics(patError*& err) {

  if (theSample == NULL) {
    err = new patErrMiscError("No sample has been loaded. Biogeme cannot estimate the parameters") ;
    WARNING(err->describe()) ;
    return ;
  }

  if (theStatistics == NULL) {
    return ;
  }

  theStatistics->bio__generateStatistics(theSample, err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  vector<pair<patString,patReal> >* theStats = theStatistics->getStatistics() ;
  for (vector<pair<patString,patReal> >::iterator i = theStats->begin() ;
       i != theStats->end() ;
       ++i) {
    theReport.addStatistic(i->first,i->second) ;
  }
}

void bioMain::setParameters(patError*& err) {
  theTrParameters.eta1 = bioParameters::the()->getValueReal("BTREta1",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.eta2 = bioParameters::the()->getValueReal("BTREta2",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.gamma2 = bioParameters::the()->getValueReal("BTRGamma2",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.beta = bioParameters::the()->getValueReal("BTRIncreaseTRRadius",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.maxIter = bioParameters::the()->getValueInt("BTRMaxIter",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.initQuasiNewtonWithTrueHessian = bioParameters::the()->getValueInt("BTRInitQuasiNewtonWithTrueHessian",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.significantDigits = bioParameters::the()->getValueInt("BTRSignificantDigits",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.usePreconditioner = bioParameters::the()->getValueInt("BTRUsePreconditioner",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.maxTrustRegionRadius = bioParameters::the()->getValueReal("BTRMaxTRRadius",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.typicalF = bioParameters::the()->getValueReal("BTRTypf",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.tolerance = bioParameters::the()->getValueReal("BTRTolerance",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.toleranceSchnabelEskow = bioParameters::the()->getValueReal("TolSchnabelEskow",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.exactHessian = bioParameters::the()->getValueInt("BTRExactHessian",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.cheapHessian = bioParameters::the()->getValueInt("BTRCheapHessian",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.initRadius = bioParameters::the()->getValueReal("BTRInitRadius",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.minRadius = bioParameters::the()->getValueReal("BTRMinTRRadius",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.armijoBeta1  = bioParameters::the()->getValueReal("BTRArmijoBeta1",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.armijoBeta2  = bioParameters::the()->getValueReal("BTRArmijoBeta2",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.stopFileName  = bioParameters::the()->getValueString("stopFileName",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.startDraws  = bioParameters::the()->getValueInt("BTRStartDraws",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.increaseDraws  = bioParameters::the()->getValueInt("BTRIncreaseDraws",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.maxGcpIter  = bioParameters::the()->getValueInt("maxGcpIter",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.fractionGradientRequired  = bioParameters::the()->getValueReal("TSFractionGradientRequired",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.expTheta  = bioParameters::the()->getValueReal("TSExpTheta",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.infeasibleCgIter = bioParameters::the()->getValueInt("BTRUnfeasibleCGIterations",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.quasiNewtonUpdate  = bioParameters::the()->getValueInt("BTRQuasiNewtonUpdate",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.kappaUbs  = bioParameters::the()->getValueReal("BTRKappaUbs",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.kappaLbs  = bioParameters::the()->getValueReal("BTRKappaLbs",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.kappaFrd  = bioParameters::the()->getValueReal("BTRKappaFrd",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  theTrParameters.kappaEpp  = bioParameters::the()->getValueReal("BTRKappaEpp",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }


  theSolvoptParameters.errorArgument = bioParameters::the()->getValueReal("solvoptErrorArgument",err) ;
 if (err != NULL) {
   WARNING(err->describe()) ;
   return ;
 }
 theSolvoptParameters.errorFunction = bioParameters::the()->getValueReal("solvoptErrorFunction",err) ;
 if (err != NULL) {
   WARNING(err->describe()) ;
   return ;
 }
 theSolvoptParameters.maxIter = bioParameters::the()->getValueInt("solvoptMaxIter",err) ;
 if (err != NULL) {
   WARNING(err->describe()) ;
   return ;
 }
 theSolvoptParameters.display = bioParameters::the()->getValueInt("solvoptDisplay",err) ;
 if (err != NULL) {
   WARNING(err->describe()) ;
   return ;
 }

 theDonlp2Parameters.epsx = bioParameters::the()->getValueReal("donlp2epsx",err) ;
 if (err != NULL) {
   WARNING(err->describe()) ;
   return ;
 }
 theDonlp2Parameters.delmin = bioParameters::the()->getValueReal("donlp2delmin",err) ;
 if (err != NULL) {
   WARNING(err->describe()) ;
   return ;
 }
 theDonlp2Parameters.smallw = bioParameters::the()->getValueReal("donlp2smallw",err) ;
 if (err != NULL) {
   WARNING(err->describe()) ;
   return ;
 }
 theDonlp2Parameters.epsdif = bioParameters::the()->getValueReal("donlp2epsdif",err) ;
 if (err != NULL) {
   WARNING(err->describe()) ;
   return ;
 }
 theDonlp2Parameters.nreset = bioParameters::the()->getValueInt("donlp2nreset",err) ;
 if (err != NULL) {
   WARNING(err->describe()) ;
   return ;
 }
 theDonlp2Parameters.stopFileName = bioParameters::the()->getValueString("stopFileName",err) ;
 if (err != NULL) {
   WARNING(err->describe()) ;
   return ;
 }

}
