//-*-c++-*------------------------------------------------------------
//
// File name : bioPrecompiledFunction.cc
// Author :    Michel Bierlaire
// Date :      Mon Apr 18 18:09:01 2011
//
//--------------------------------------------------------------------

#include <iterator>
#include "bioPrecompiledFunction.h"
#include "trParameters.h"
#include "bioLiteralRepository.h"
#include "bioParameters.h"
#include "bioIteratorInfoRepository.h"
#include "bioExpression.h"
#include "patErrNullPointer.h"
#include "bioExpressionRepository.h"
#include "patLargeNumbers.h"
#include <pthread.h> 
void *computeFunction( void *ptr );
void *computeFunctionAndGradient( void *ptr );



bioPrecompiledFunction::bioPrecompiledFunction(bioExpressionRepository* rep, 
					       patULong expId,
					       patError*& err) 
: theExpressionRepository(bioParameters::the()->getValueInt("numberOfThreads")), 
  theExpressionId(expId),
  bhhh(theTrParameters,bioLiteralRepository::the()->getNumberOfEstimatedParameters()), 
  hessian(theTrParameters,bioLiteralRepository::the()->getNumberOfEstimatedParameters()), 
  threadGrad(bioParameters::the()->getValueInt("numberOfThreads"),
	     trVector(bioLiteralRepository::the()->getNumberOfEstimatedParameters())), 
  threadBhhh(bioParameters::the()->getValueInt("numberOfThreads"),
	     trHessian(theTrParameters,bioLiteralRepository::the()->getNumberOfEstimatedParameters())), 
  threadSuccess(bioParameters::the()->getValueInt("numberOfThreads"),patFALSE),
  threadHessian(bioParameters::the()->getValueInt("numberOfThreads"),
		trHessian(theTrParameters,
			  bioLiteralRepository::the()->getNumberOfEstimatedParameters() )), 
  threadChildren(bioParameters::the()->getValueInt("numberOfThreads")),
  threadSpans(bioParameters::the()->getValueInt("numberOfThreads")) 
{

  DEBUG_MESSAGE("Repository of size: " << rep->getNbrOfExpressions()) ;
  bioExpression* theExpression = rep->getExpression(theExpressionId) ;
  if (theExpression != NULL) {
    DEBUG_MESSAGE("Funct: " << *theExpression) ;
  }

  patULong nThreads = bioParameters::the()->getValueInt("numberOfThreads") ;
  if (theExpression->containsAnIterator() && nThreads > 1 && !theExpression->isSumIterator()) {
    err = new patErrMiscError("Multithreading can be used only if the expression is a sum over the sample. If you want to work with this expression, reduce the number of threads to 1: BIOGEME_OBJECT.PARAMETERS['numberOfThreads'] = \"1\"") ;
    WARNING(err->describe()) ;
    return  ;
  }

  bhhh.setToZero() ;
  hessian.setToZero() ;
  betaIds = bioLiteralRepository::the()->getBetaIds(patFALSE, err)  ;

  
  for (patULong i = 0 ; i < theExpressionRepository.size() ; ++i) {
    if (rep != NULL) {
      bioExpressionRepository* theCopy = rep->getCopyForThread(i,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
      }
      theCopy->check(err) ;
      if (err != NULL) {
	DEBUG_MESSAGE("Original: " << rep->printList()) ;
	DEBUG_MESSAGE("Copy: " << theCopy->printList()) ;
	WARNING(err->describe()) ;
      }
      theCopy->getExpression(theExpressionId) ;
      // theExpr->prepareDerivatives(betaIds,err) ;
      // if (err != NULL) {
      // 	WARNING(err->describe()) ;
      // }

      theExpressionRepository[i] = theCopy ;
    }
    else {
      theExpressionRepository[i] = NULL ;
    }
  }

  threads = new pthread_t[nThreads];
  input = new bioThreadArg[nThreads];

}

bioPrecompiledFunction::~bioPrecompiledFunction() {

}

patReal bioPrecompiledFunction::getFunction(trVector* x,
			       patBoolean* success,
			       patError*& err) {


  short nThreads = bioParameters::the()->getValueInt("numberOfThreads") ;
  bioLiteralRepository::the()->setBetaValues(*x,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  trVector theBetas = bioLiteralRepository::the()->getBetaValues() ; 
  patReal logLike = 0.0 ;
  
  
  for(int i = 0; i < nThreads; ++i) {
    input[i].threadId = i ;
    input[i].grad = NULL ;
    input[i].hessian = NULL ;
    input[i].bhhh = NULL ;
    input[i].success = &(threadSuccess[i]) ;
    input[i].sample = sample ;
    input[i].result = 0.0 ;
    threadSpans[i] = bioIteratorInfoRepository::the()->getSublist(i,err) ; 
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patReal();
    }
    input[i].subsample = &(threadSpans[i]) ;
    input[i].theExpressionRepository = theExpressionRepository[i] ;
    input[i].theExpressionId = theExpressionId ;

    if(pthread_create( &threads[i], NULL, computeFunction, (void*) &input[i])) {
      err = new patErrMiscError("Error creating Threads! Exit!") ;
      WARNING(err->describe()) ;
      return 0.0;
    }
  }

 logLike = 0.0 ;
 patBoolean maxreal(patFALSE) ;
 for(int i=0; i<nThreads; ++i) {
   pthread_join( threads[i], NULL);
   if (!maxreal) {
     if (input[i].result == -patMaxReal) {
       logLike = patMaxReal ;
       maxreal = patTRUE ;
     }
     else {
       logLike -= input[i].result ;
     }
   }
 }
 
 if (!finite(logLike)) {
   return patMaxReal ;
 }
 return logLike ;
}
  


trVector* bioPrecompiledFunction::getGradient(trVector* x,
				 trVector* grad,
				 patBoolean* success,
				 patError*& err) {
  *success = patTRUE ;
  if (grad == NULL) {
    err = new patErrNullPointer("trVector") ;
    WARNING(err->describe()) ;
    return NULL ;
  }
  getFunctionAndGradient(x,grad,success,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL  ;
  }
  return grad ;
}

patReal bioPrecompiledFunction::getFunctionAndGradient(trVector* x,
					  trVector* grad,
					  patBoolean* success,
					  patError*& err) {



  if (grad == NULL) {
    err = new patErrNullPointer("trVector") ;
    WARNING(err->describe()) ;
    return patReal() ;
  }

  short nThreads = bioParameters::the()->getValueInt("numberOfThreads") ;
  *success = patTRUE ;
  bhhh.setToZero() ;
  bioLiteralRepository::the()->setBetaValues(*x,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  trVector theBetas = bioLiteralRepository::the()->getBetaValues() ; 
  patReal logLike = 0.0 ;
  
  fill(grad->begin(),grad->end(),0.0) ; 
  
  for(int i=0; i<nThreads; i++) {
    input[i].threadId = i ;
    fill(threadGrad[i].begin(),threadGrad[i].end(),0.0) ;
    input[i].grad = &(threadGrad[i]) ;
    threadHessian[i].setToZero() ;    
    input[i].hessian = &(threadHessian[i]) ;
    threadBhhh[i].setToZero() ;    
    input[i].bhhh = &(threadBhhh[i]) ;
    input[i].success = &(threadSuccess[i]) ;
    input[i].sample = sample ;
    input[i].result = 0.0 ;
    input[i].literalIds = betaIds ;
    threadSpans[i] = bioIteratorInfoRepository::the()->getSublist(i,err) ;  
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patReal();
    }
    input[i].subsample = &(threadSpans[i]) ;
    input[i].theExpressionRepository = theExpressionRepository[i] ;
    input[i].theExpressionId = theExpressionId ;

    if(pthread_create( &threads[i], NULL, computeFunctionAndGradient, (void*) &input[i])) {
      err = new patErrMiscError("Error creating Threads! Exit!") ;
      WARNING(err->describe()) ;
      return 0.0;
    }
  }
  
  void *tmp;
  
  patBoolean maxreal(patFALSE) ;
  logLike = 0.0 ; 
  for(int i=0; i<nThreads; ++i) {
    pthread_join( threads[i], &tmp);
    if (!maxreal) {
      if (input[i].result == -patMaxReal) {
	logLike = patMaxReal ;
	maxreal = patTRUE ;
      }
      else {
	logLike -= input[i].result ;
	(*grad) -= threadGrad[i] ;
	bhhh.add(1.0,threadBhhh[i],err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return 0.0;
	}
      }
    }
  }
	
  if (!finite(logLike)) {
    return patMaxReal ;
  }
  return logLike ;

}


trHessian* bioPrecompiledFunction::computeHessian(patVariables* x,
				     trHessian& hessian,
				     patBoolean* success,
				     patError*& err) {
  return computeFinDiffHessian(x,hessian,success,err) ;
}

trHessian* bioPrecompiledFunction::getCheapHessian(trVector* x,
				      trHessian& hessian,
				      patBoolean* success,
				      patError*& err) {

  hessian = bhhh ;
  *success = patTRUE ;
  return &bhhh ;
}

patBoolean bioPrecompiledFunction::isCheapHessianAvailable() {
  return patTRUE ;
}

trVector bioPrecompiledFunction::getHessianTimesVector(trVector* x,
					  const trVector* v,
					  patBoolean* success,
					  patError*& err) {
  return trVector() ;
}

patBoolean bioPrecompiledFunction::isGradientAvailable(patError*& err) const {
  return patTRUE ;
}

patBoolean bioPrecompiledFunction::isHessianAvailable(patError*& err)  const {
  return patTRUE ;
}

patBoolean bioPrecompiledFunction::isHessianTimesVectorAvailable(patError*& err) const {
  return patFALSE ;
}

unsigned long bioPrecompiledFunction::getDimension(patError*& err) const {
  return bioLiteralRepository::the()->getNumberOfEstimatedParameters() ;
}

patBoolean bioPrecompiledFunction::isUserBased() const {
  return patTRUE ;
}


void *computeFunction(void* fctPtr) {
  patError* err(NULL) ;
  bioThreadArg *input = (bioThreadArg *) fctPtr;

  const bioSample* sample = input->sample ;
  bioIteratorSpan __theThreadSpan = *input->subsample ; 
  bioExpressionRepository* theRep = input->theExpressionRepository ;
  bioExpression* theExpression = theRep->getExpression(input->theExpressionId) ;
  if (theExpression == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return NULL;
  }

  theExpression->setSample(sample) ;
  theExpression->setThreadSpan(__theThreadSpan) ;
  
  // DEBUG_MESSAGE("THREAD " << input->threadId <<" LIST OF EXPRESSIONS") ;
  // cout << bioExpressionRepository::the()->printList() << endl ;

  input->result = theExpression->getValue(patFALSE,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
  }
  return NULL ;
}

void *computeFunctionAndGradient( void *ptr ) {
  patError* err(NULL) ;
  bioThreadArg *input = (bioThreadArg *) ptr;
  const bioSample* sample = input->sample ;
  bioIteratorSpan __theThreadSpan = *input->subsample ; 
  vector<patULong> literalIds = input->literalIds ;
  patULong exprId = input->theExpressionId ;
  bioExpressionRepository* theRepository = input->theExpressionRepository ;
  bioExpression* theExpression = theRepository->getExpression(exprId) ;
  if (theExpression == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return NULL;
  }
  
  theExpression->setSample(sample) ;
  theExpression->setThreadSpan(__theThreadSpan) ;
  
  //bioFunctionAndDerivatives result = theExpression->getFunctionAndGradient(err) ;
  bioFunctionAndDerivatives* result = theExpression->getNumericalFunctionAndGradient(literalIds,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  
  if (result->theGradient.size() != input->grad->size()) {
    stringstream str ;
    str << "Incompatible sizes: " << result->theGradient.size() << " <> " << input->grad->size() ;
    err = new patErrMiscError(str.str());
    WARNING(err->describe());
    return NULL ;
  }
  
  input->result = result->theFunction ;
  for (patULong i = 0 ; i < result->theGradient.size() ; ++i) {
    (*input->grad)[i] = result->theGradient[i] ;
  }

  trHessian* bhhh = theExpression->getBhhh() ;
  if (bhhh != NULL) {
    
    input->bhhh->set(*bhhh,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
  }
  return NULL ;
}
